use serde::{Serialize, Deserialize};
use crate::{
    models::item::{SkillType, Item, Pronouns},
    models::consent::ConsentType,
    message_handler::user_commands::{UResult, VerbContext},
    static_content::{
        room::Direction,
        species::BodyPart,
    },
};
use once_cell::sync::OnceCell;
use std::collections::BTreeMap;
use rand::seq::SliceRandom;
use async_trait::async_trait;

mod fangs;
mod whip;
mod blade;
mod trauma_kit;
mod corp_licence;
mod lock;
mod meat;
pub mod head_armour;
pub mod torso_armour;
pub mod lower_armour;

pub type AttackMessageChoice = Vec<Box<dyn Fn(&Item, &Item, bool) -> String + 'static + Sync + Send>>;
pub type AttackMessageChoicePart = Vec<Box<dyn Fn(&Item, &Item, &BodyPart, bool) -> String + 'static + Sync + Send>>;

pub struct SkillScaling {
    pub skill: SkillType,
    pub min_skill: f64,
    pub mean_damage_per_point_over_min: f64
}

#[allow(unused)]
#[derive(Eq,Ord,Clone,PartialEq,PartialOrd)]
pub enum DamageType {
    Beat,
    Slash,
    Pierce,
    Shock,
    Bullet
}

pub struct WeaponAttackData {
    pub start_messages: AttackMessageChoice,
    pub success_messages: AttackMessageChoicePart,
    pub mean_damage: f64,
    pub stdev_damage: f64,
    pub base_damage_type: DamageType,
    pub other_damage_types: Vec<(f64, DamageType)>, // Allocation fractions.
    pub skill_scaling: Vec<SkillScaling>,
}

impl Default for WeaponAttackData {
    fn default() -> Self {
        Self {
            start_messages:
            vec!(Box::new(|attacker, victim, exp| format!(
                "{} makes an attack on {}",
                &attacker.display_for_sentence(exp, 1, true),
                &victim.display_for_sentence(exp,1, false)))),
            success_messages:
            vec!(Box::new(|attacker, victim, part, exp|
                          format!("{}'s attack on {} hits {} {}",
                                  &attacker.display_for_sentence(exp, 1, true),
                                  &victim.display_for_sentence(exp, 1, false),
                                  &victim.pronouns.possessive,
                                  part.display(victim.sex.clone())
                          ))),
            mean_damage: 1.0,
            stdev_damage: 2.0,
            base_damage_type: DamageType::Slash,
            other_damage_types: vec!(),
            skill_scaling: vec!()
        }
    }
}

pub struct WeaponData {
    pub uses_skill: SkillType,
    pub raw_min_to_learn: f64,
    pub raw_max_to_learn: f64,
    pub normal_attack: WeaponAttackData,
}

impl Default for WeaponData {
    fn default() -> Self {
        Self {
            uses_skill: SkillType::Blades,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 15.0,
            normal_attack: Default::default(),
        }
    }
}

pub struct ChargeData {
    pub max_charges: u8,
    pub charge_name_prefix: &'static str,
    pub charge_name_suffix: &'static str,
}

impl Default for ChargeData {
    fn default() -> Self {
        Self {
            max_charges: 1,
            charge_name_prefix: "charge",
            charge_name_suffix: "",
        }
    }
}

pub enum UseEffect {
    // messagef takes player, item used, target as the 3 parameters. Returns (explicit, non explicit) message.
    BroadcastMessage { messagef: Box<dyn Fn(&Item, &Item, &Item) -> (String, String) + Sync + Send>},
    // skill_multiplier is always positive - sign flipped for crit fails.
    ChangeTargetHealth { delay_secs: u64, base_effect: i64, skill_multiplier: f64,
                         max_effect: i64, message: Box<dyn Fn(&Item) -> (String, String) + Sync + Send> },
}

pub struct UseData {
    pub uses_skill: SkillType,
    pub diff_level: f64,
    pub crit_fail_effects: Vec<UseEffect>,
    pub fail_effects: Vec<UseEffect>,
    pub success_effects: Vec<UseEffect>,
    pub errorf: Box<dyn Fn(&Item, &Item) -> Option<String> + Sync + Send>,
    pub task_ref: &'static str,
    pub needs_consent_check: Option<ConsentType>,
}

impl Default for UseData {
    fn default() -> Self {
        Self {
            uses_skill: SkillType::Medic,
            diff_level: 10.0,
            crit_fail_effects: vec!(),
            fail_effects: vec!(),
            success_effects: vec!(),
            errorf: Box::new(|_it, _target| None),
            task_ref: "set me",
            needs_consent_check: None,
        }
    }
}

pub struct SoakData {
    pub min_soak: f64,
    pub max_soak: f64,
    pub damage_probability_per_soak: f64,
}

pub struct WearData {
    pub covers_parts: Vec<BodyPart>,
    pub thickness: f64,
    pub dodge_penalty: f64,
    pub soaks: BTreeMap<DamageType, SoakData>,
}

#[async_trait]
pub trait WriteHandler {
    async fn write_cmd(&self, ctx: &mut VerbContext, player: &Item, on_what: &Item, write_what: &str) -> UResult<()>;
}

#[async_trait]
pub trait ArglessHandler {
    async fn cmd(&self, ctx: &mut VerbContext, player: &Item, what: &Item) -> UResult<()>;
}

#[async_trait]
pub trait InstallHandler {
    async fn install_cmd(&self, ctx: &mut VerbContext, player: &Item, what: &Item, room: &Item, direction: &Direction) -> UResult<()>;
    async fn uninstall_cmd(&self, ctx: &mut VerbContext, player: &Item, what: &Item, room: &Item, direction: &Direction) -> UResult<()>;
}

pub struct PossessionData {
    pub weapon_data: Option<WeaponData>,
    pub display: &'static str,
    pub display_less_explicit: Option<&'static str>,
    pub details: &'static str, 
    pub details_less_explicit: Option<&'static str>,
    pub aliases: Vec<&'static str>,
    pub max_health: u64,
    pub charge_data: Option<ChargeData>,
    pub use_data: Option<UseData>,
    pub becomes_on_spent: Option<PossessionType>,
    pub weight: u64,
    pub install_handler: Option<&'static (dyn InstallHandler + Sync + Send)>,
    pub lockcheck_handler: Option<&'static (dyn ArglessHandler + Sync + Send)>,
    pub sign_handler: Option<&'static (dyn ArglessHandler + Sync + Send)>,
    pub write_handler: Option<&'static (dyn WriteHandler + Sync + Send)>,
    pub can_butcher: bool,
    pub wear_data: Option<WearData>
}

impl Default for PossessionData {
    fn default() -> Self {
        Self {
            weapon_data: None,
            display: "Thingy",
            display_less_explicit: None,
            details: "A generic looking thing",
            details_less_explicit: None,
            aliases: vec!(),
            max_health: 10,
            weight: 100,
            charge_data: None,
            becomes_on_spent: None,
            use_data: None,
            install_handler: None,
            lockcheck_handler: None,
            sign_handler: None,
            write_handler: None,
            can_butcher: false,
            wear_data: None,
        }
    }
}

impl WeaponAttackData {
    pub fn start_message(
        &self,
        attacker: &Item, victim: &Item, explicit_ok: bool) -> String {
        let mut rng = rand::thread_rng();
        self.start_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    pub fn success_message(
        &self, attacker: &Item, victim: &Item,
        part: &BodyPart, explicit_ok: bool
    ) -> String {
        let mut rng = rand::thread_rng();
        self.success_messages[..].choose(&mut rng).map(
            |f| f(attacker, victim, part, explicit_ok)).unwrap_or(
            "No message defined yet".to_owned())
    }

    
}

#[derive(Serialize, Deserialize, Clone, Debug, PartialEq, Eq, PartialOrd, Ord)]
pub enum PossessionType {
    // Special values that substitute for possessions.
    Fangs, // Default weapon for certain animals
    // Real possessions from here on:
    // Armour
    RustyMetalPot,
    HockeyMask,
    LeatherJacket,
    LeatherPants,
    // Weapons: Whips
    AntennaWhip,
    // Weapons: Blades
    ButcherKnife,
    // Medical
    MediumTraumaKit,
    EmptyMedicalBox,
    // Corporate
    NewCorpLicence,
    CertificateOfIncorporation,
    // Security
    Scanlock,
    // Crafting
    Steak,
    AnimalSkin,
    SeveredHead,
    
}

impl Into<Item> for PossessionType {
    fn into(self) -> Item {
        let possession_dat = possession_data().get(&self).unwrap();
        Item {
            item_type: "possession".to_owned(),
            possession_type: Some(self.clone()),
            display: possession_dat.display.to_owned(),
            display_less_explicit: possession_dat.display_less_explicit.map(|d| d.to_owned()),
            details: Some(possession_dat.details.to_owned()),
            details_less_explicit: possession_dat.details_less_explicit.map(|d| d.to_owned()),
            aliases: possession_dat.aliases.iter().map(|al| (*al).to_owned()).collect(),
            health: possession_dat.max_health,
            weight: possession_dat.weight,
            pronouns: Pronouns {
                is_proper: false,
                ..Pronouns::default_inanimate()
            },
            charges: possession_dat.charge_data.as_ref()
                .map(|cd| cd.max_charges).unwrap_or(0),
            ..Default::default()
        }
    }
}

pub fn fist() -> &'static WeaponData {
    static FIST_WEAPON: OnceCell<WeaponData> = OnceCell::new();
    FIST_WEAPON.get_or_init(|| {
        WeaponData {
            uses_skill: SkillType::Fists,
            raw_min_to_learn: 0.0,
            raw_max_to_learn: 2.0,
            normal_attack: WeaponAttackData {
                start_messages: vec!(
                    Box::new(|attacker, victim, exp|
                             format!("{} swings at {} with {} fists",
                                     &attacker.display_for_sentence(exp, 1, true),
                                     &victim.display_for_sentence(exp, 1, false),
                                     &attacker.pronouns.possessive
                             )
                    )
                ),
                success_messages: vec!(
                    Box::new(|attacker, victim, part, exp|
                             format!("{}'s fists smash into {}'s {}",
                                     &attacker.display_for_sentence(exp, 1, true),
                                     &victim.display_for_sentence(exp, 1, false),
                                     &part.display(victim.sex.clone())
                             )
                    )
                ),
                ..Default::default()
            },
            ..Default::default()
        }
    })
}

pub fn possession_data() -> &'static BTreeMap<PossessionType, &'static PossessionData> {
    static POSSESSION_DATA: OnceCell<BTreeMap<PossessionType, &'static PossessionData>> = OnceCell::new();
    use PossessionType::*;
    &POSSESSION_DATA.get_or_init(|| {
        vec!(
            (Fangs, fangs::data())
        ).into_iter()
            .chain(whip::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(blade::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(trauma_kit::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(corp_licence::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(lock::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(meat::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(head_armour::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(torso_armour::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .chain(lower_armour::data().iter().map(|v| ((*v).0.clone(), &(*v).1)))
            .collect()
    })
}

pub fn can_butcher_possessions() -> &'static Vec<PossessionType> {
    static RELEVANT: OnceCell<Vec<PossessionType>> = OnceCell::new();
    &RELEVANT.get_or_init(|| {
        possession_data().iter()
            .filter_map(|(pt, pd)| if pd.can_butcher { Some(pt.clone()) } else { None })
            .collect()
    })
}

#[cfg(test)]
mod tests {
    use super::*;
    #[test]
    fn other_damage_types_add_to_less_than_one() {
        for (_pt, pd) in possession_data().iter() {
            if let Some(weapon_data) = pd.weapon_data.as_ref() {
                let tot: f64 = weapon_data.normal_attack.other_damage_types.iter().map(|v|v.0).sum();
                assert!(tot >= 0.0);
                assert!(tot < 1.0);
            }
        }
    }
}
